/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2017/18      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : tutorial 5
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: merge sort
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>
#include <iomanip>
#include <vector>
#include <thread>
#include <cstdlib>
#include <chrono>
#include <ctime>
#include <cassert>
#include <algorithm>

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge step of merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
merge( T* data, T* last, int nhalf, T* scratch) {

  int* data1 = data;
  int* last1 = data + nhalf;
  int* data2 = data + nhalf;
  int* last2 = last;

  // merge
  T* scratchref = scratch;
  while( data1 != last1 && data2 != last2) {
    T*& which = (*data1 < *data2) ? data1 : data2;
    *scratch++ = std::move( *which++);
  }
  // copy remainder (obviously, only one will have an effect)
  std::copy( data1, last1, scratch);
  std::copy( data2, last2, scratch);

  memcpy( data, scratchref, (last-data) * sizeof(T));
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
mergesort( T* data, T* last, T* scratch ) {

  const int nmin = 2;
  int nhalf = (last - data) / 2;

  if( nhalf < nmin) {
    std::sort( data, last);
    return;
  }

  // recursion
  mergesort( data, data + nhalf, scratch);
  mergesort( data + nhalf, last, scratch);

  // merge step
  merge( data, last, nhalf, scratch);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
mergesortParallel( T* data, T* last, T* scratch, const int num_threads) {

  int chunk_size = (last - data) / num_threads;
  assert( 0 == (last - data) % num_threads);
  assert( 0 == num_threads % 2);

  // run merge sort on chunks
  std::vector< std::thread > threads;
  for( int tid = 0; tid < num_threads; ++tid) {
    threads.push_back( std::thread( mergesort<T>, data + tid*chunk_size, data + (tid+1)*chunk_size,
                                                  scratch + tid*chunk_size));
  }

  for( auto& th : threads) {
    th.join();
  }

  // merge chunks, traversing the tree upwards towards the root
  int active_chunks = num_threads;
  while( active_chunks > 1) {

    // merge adjacent nodes in the tree
    for( int i = 0; i < active_chunks; i+=2) {
      merge( data + i * chunk_size, data + (i+2) * chunk_size, chunk_size,
             scratch + i * chunk_size);
    }

    chunk_size *= 2;
    active_chunks /= 2;
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  std::srand( std::time(0));

  // allocate memory
  int n = 4096;
  int* data = (int*) malloc( n * sizeof(int));
  int* scratch = (int*) malloc( n * sizeof(int));
  std::generate( data, data + n, std::rand);

  // reference solution
  int* dataref = (int*) malloc( n * sizeof(int));
  memcpy( dataref, data, n * sizeof(int));
  std::sort( dataref, dataref+n);

#if 0
  // run merge sort
  mergesort( data, data+n, scratch);
#endif

#if 1
  // run parallel merge sort
  unsigned int num_threads = std::thread::hardware_concurrency();
  std::cout << "num_threads = " << num_threads << std::endl;
  mergesortParallel( data, data+n, scratch, num_threads);
#endif

  // check correctness of result
  bool correct = true;
  for( int i = 0; i < n; ++i) {
    correct &= (dataref[i] == data[i]);
  }
  std::cout << "Mergesort " << ((correct) ? "succeeded." : "failed.") << std::endl;

  // clean up
  free( data);
  free( scratch);
  free( dataref);

  return EXIT_SUCCESS;
}
